@@ -1,5 +1,17 @@
+
+#include <stdio.h>
+#include <stdlib.h>
+
 #include "leak_detector.h"
 
+#define CHECK_OUT_OF_MEMORY(p) \
+    do { \
+        if (NULL == (p)) { \
+            fprintf(stderr, "Out of memory\n"); \
+            exit(69); \
+        } \
+    } while (0)
+
 int main(void) {
     char *p[10];
     int i;
@@ -10,6 +22,7 @@ int main(void) {
                 /* FIXME: This is not ideal.                            */
     for (i = 0; i < 10; ++i) {
         p[i] = (char*)malloc(sizeof(int)+i);
+        CHECK_OUT_OF_MEMORY(p[i]);
     }
     CHECK_LEAKS();
     for (i = 1; i < 10; ++i) {
@@ -17,6 +30,7 @@ int main(void) {
     }
     for (i = 0; i < 9; ++i) {
         p[i] = (char*)malloc(sizeof(int)+i);
+        CHECK_OUT_OF_MEMORY(p[i]);
     }
     CHECK_LEAKS();
     CHECK_LEAKS();
